################################################################################
#   Copyright 2011 Chris Statzer and David Reichard                            #
#                                                                              #
#   Licensed under the Apache License, Version 2.0 (the "License");            #
#   you may not use this file except in compliance with the License            #
#   You may obtain a copy of the License at                                    #
#                                                                              #
#       http://www.apache.org/licenses/LICENSE-2.0                             #
#                                                                              #
#   Unless required by applicable law or agreed to in writing, software        #
#   distributed under the License is distributed on an "AS IS" BASIS,          #
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.   #
#   See the License for the specific language governing permissions and        #
#   limitations under the License.                                             #
################################################################################

import pyglet
import collision
from vec2d import vec2d
from random import randint

from PIL import Image

class Entity(pyglet.sprite.Sprite):
    def __init__(self, image, x=0, y=0, radius=1.0, blend_src=770, blend_dest=771, batch=None, group=None, usage='dynamic'):
        pyglet.sprite.Sprite.__init__(self, image, x, y, blend_src, blend_dest, batch, group, usage)
        self.pos = vec2d(x, y)
        self.velocity = vec2d(0, 0)
        self.radius = radius
        self.old_pos = vec2d(0, 0) # used for collision checking purposes
        
    def collides_with(self, other):
        """
        Determine if this entity is colliding with another.
        
        Returns True if collision is detected, False if otherwise.
        """

        return collision.circle_circle_intersect(self.pos, self.radius, other.pos, other.radius)
        
    def check_walls(self, terrain, game_map, bouncing=False):
        # check for collisions against wall
        # if one is found, get new position, to allow "sliding"
        # against them
        
        #bouncing is supported also for molecules and such
        
        # check distance from self to terrain, if more than terrain's width
        # then there is no need for checking collision for terrain walls
        
        #if not self.pos.get_dist_sqrd((terrain.x, terrain.y)) > (self.radius + terrain.image.width) ** 2:
        if True: #blah
            lines = terrain.walls
            i = 0
            while i < len(lines):
                p1 = vec2d(lines[i], lines[i+1])
                p2 = vec2d(lines[i+2], lines[i+3])
                line = p2 - p1
                normal = line.perpendicular_normal()
                vec = (self.x, self.y) + self.radius * normal
                intersects = collision.line_intersect(self.old_pos, vec, p1, p2)
                if not intersects is None:
                    if bouncing:
                        self.velocity = 2 * ((self.velocity*-1).dot(normal)) * normal - (self.velocity * -1)
                      #  self.position += self.velocity * 0.1  #move out a bit to avoid floating point errors
                    else:
                        new_pos = collision.calc_new_pos(self.old_pos, vec, p1, p2)
                        new_pos -= self.radius * normal
                        self.x = new_pos.x
                        self.y = new_pos.y
                elif collision.circle_point_intersect(self.position, self.radius, p1):
                    direction = (self.position - p1)
                    length = direction.length
                    move_by_distance = self.radius - length
                    new_pos = self.position + direction.normalized() * move_by_distance
                    self.x = new_pos.x
                    self.y = new_pos.y
                i += 4
                
        # check against game map boundaries
        #right
        if self.x + self.radius > game_map.width:
            self.x -= self.x + self.radius - game_map.width
            if bouncing:
                self.velocity.x *= -1
            
        #left
        elif self.x - self.radius < 0:
            self.x += abs(self.x - self.radius)
            if bouncing:
                self.velocity.x *= -1
        #top
        elif self.y + self.radius > game_map.height:
            self.y -= self.y + self.radius - game_map.height
            if bouncing:
                self.velocity.y *= -1
        #bottom
        elif self.y - self.radius < 0:
            self.y += abs(self.y - self.radius)
            if bouncing:
                self.velocity.y *= -1
                
    def update(self, dt):
        #update our vec2d vector and old_pos(for collision checking purposes)
        self.old_pos = vec2d(self.x, self.y)
        self.pos = vec2d(self.position)

###############################################################################
################################################################ Cell ######### 
###############################################################################

class Cell (Entity):

    def __init__(self, x, y):
        image = pyglet.resource.image('cell.png')        
        image.anchor_x = image.width / 2
        image.anchor_y = image.height / 2
        Entity.__init__(self, image, x, y, image.width // 2 * 0.8)
        self.x, self.y = x, y
        self.velocity = vec2d(0.0, 0.0)
        self.max_velocity = 150 # max velocity of 150 game units/sec
        self.acceleration_factor = 300  # acceleration of 300 game units/sec
        self.friction = 0.995
        self.moving = vec2d(0, 0) # which direction we are accelerating in

    def update(self, dt, game_map):
        Entity.update(self, dt)
        acceleration = self.moving * self.acceleration_factor * dt
        
        self.velocity *= self.friction
        self.velocity += acceleration
        
        #going to save on computation by avoiding square root here
        if self.velocity.get_length_sqrd() > self.max_velocity**2:
            self.velocity.length = self.max_velocity
        
        # move the cell towards it's velocity
        self.x += self.velocity[0] * dt
        self.y += self.velocity[1] * dt

        if self.velocity != (0,0):
            for terrain in game_map.terrain:
                self.check_walls(terrain, game_map)
            self.rotation += 1
        
###############################################################################
################################################################ Molecule ##### 
###############################################################################

class Molecule (Entity):

    def __init__(self, image, x, y, nums):
        image.anchor_x = image.width / 2
        image.anchor_y =  image.height / 2
        Entity.__init__(self, image, x, y, image.width // 2 * 0.8)
        self.numbers = nums
        self.velocity = vec2d(randint(-100,100),randint(-100,100))
        self.col_timer = 0

    def update(self, dt, game_map):
        self.col_timer += dt
        
        self.rotation += 45 *dt

        # move the cell towards it's velocity
        self.x += self.velocity[0] * dt
        self.y += self.velocity[1] * dt
        
        if self.col_timer >= 0.3:
            self.col_timer = 0        
            if self.velocity != (0,0):
                for terrain in game_map.terrain:
                    self.check_walls(terrain, game_map, True)
        
        Entity.update(self, dt)

    def kill(self):
        self.spawner.molecules.remove(self)

###############################################################################
############################################################ MoleculeFactory ##
###############################################################################

class MoleculeFactory (object):
    """
        Loads all the molecule assets and provides an interface to get a 
        weighted pseudo random molecule. make_molecule takes a level range
        where the levels are [english, roman, arabic, hebrew] or (0,3)
    """
    def __init__(self):
        self.numbers = Image.open('assets/number_sheet.png')
        self.molecules = [ Image.open('assets/molecule1.png'),
                           Image.open('assets/molecule2.png'),
                           Image.open('assets/molecule3.png'),
                           Image.open('assets/molecule4.png'),
            ]

    def make_molecule(self, lang_range, x=200, y=200):
        ## define the combos so we can make sure to spawn an above average
        ## ammount of them.
        nines = [(3,3,3), (2,2,5), (1,2,6), (1,1,7)]
        bonus = [(6,3,9), (5,4,9), (7,2,9), (8,1,9)]
        if randint(1,100) <= 50:
            if randint(1,100) <= 50:
                n = nines[randint(0,3)]
            else:
                n = bonus[randint(0,3)]
        else:
            n = [randint(1,5), randint(1,5), randint(1,5)]
        ## randomize the language for difficulty
        l = randint(*lang_range)
        image = Image.new("RGBA", (96,84))
        image.paste(self.molecules[randint(0,3)])
        number = self.numbers.crop((n[0]*36, l*36, n[0]*36+36, l*36 +36))
        image.paste(number,(29, 2), number)
        number = self.numbers.crop((n[1]*36, l*36, n[1]*36+36, l*36 +36))
        image.paste(number,(-2, 48), number)
        number = self.numbers.crop((n[2]*36, l*36, n[2]*36+36, l*36 +36))
        image.paste(number,(58, 48), number)
        ## make the pyglet image and transfer the PIL data to it.
        pyglet_image = pyglet.image.create(96,84)
        pyglet_image.set_data("RGBA", -96*4, image.tostring())
        ## build a molecule passing along the numbers in a tuple.
        return Molecule(pyglet_image, x, y, n)
