################################################################################
#   Copyright 2011 Chris Statzer and David Reichard                            #
#                                                                              #
#   Licensed under the Apache License, Version 2.0 (the "License");            #
#   you may not use this file except in compliance with the License            #
#   You may obtain a copy of the License at                                    #
#                                                                              #
#       http://www.apache.org/licenses/LICENSE-2.0                             #
#                                                                              #
#   Unless required by applicable law or agreed to in writing, software        #
#   distributed under the License is distributed on an "AS IS" BASIS,          #
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.   #
#   See the License for the specific language governing permissions and        #
#   limitations under the License.                                             #
################################################################################

"""
The game state.

THIS FILE IS FULL OF http://www.youtube.com/watch?v=GA69pmhrBiE

"""

from random import randint
import json

import pyglet
from pyglet.gl import *
from pyglet.window import key

from state import State
from vec2d import vec2d
from bubble import PositionalBubbleEmitter
from entity import Entity, Cell, Molecule, MoleculeFactory
from camera import Camera
from terrain import Terrain


###############################################################################
################################################################ Spawner ###### 
###############################################################################

class Spawner (object):
    """
        Will be updated by the main loop. These spawners should randonly
        spawn a molecule and track the ones it has spawned. It should
        accept a limit to how many it may have active. 
    """
    def __init__(self, points):
        self.points = points
        self.level = 3
        self.molecules = []
        self.mf = MoleculeFactory()
        self.max_spawns = 21
        

    def update(self, dt):
        if len(self.molecules) < self.max_spawns:
            s = self.points[randint(0, len(self.points))-1]
            m = self.mf.make_molecule((0, self.level), s[0], s[1])
            m.spawner = self
            self.molecules.append(m)


###############################################################################
################################################################ GameMap ###### 
###############################################################################

class GameMap (object):
    """
        Load a json encoded map from a file and populate the
        game data.
    """
    def __init__(self, filename):
        self.debug = False
        ## Load the ground and grass graphics
        self.ground = pyglet.resource.image("ground.png")
        self.grass_image = pyglet.resource.image("grass.png")
        ## Load the level data
        map_data = json.load(open('assets/%s.map' % filename))
        self.terrain = []
        for t in map_data['terrain']:
            self.terrain.append(Terrain(t[0], t[1], t[2]))
        self.grass = []
        for g in map_data['grass']:
            self.grass.append(g)
        self.width, self.height = map_data["size"][0]
        spawners = []
        for s in map_data['spawners']:
            spawners.append(s)
        self.spawner = Spawner(spawners)

    def draw(self):
        self.ground.blit(0,0)
        for t in self.terrain:
            t.draw(debug=self.debug)
        for g in self.grass:
            self.grass_image.blit(g[0], g[1])


###############################################################################
################################################################ ScoreWidget ## 
###############################################################################

class ScoreWidget (pyglet.text.Label):
    """
        The UI display for the score.
    """
    def __init__(self, window):
        self.score = 0
        pyglet.text.Label.__init__(self, text='Score: 0', font_size=20, 
                                    x=window.width-300, y=window.height-25)

    def add(self, n):
        """
            Add n points to the score.
        """
        self.score += n
        self.text = "Score: %d" % self.score

    def get_score(self):
        return self.score


###############################################################################
################################################################  Playing #####
###############################################################################

class Playing(State):
    def __init__(self,window, *args, **kwargs):
        State.__init__(self, window, *args, **kwargs)
        self.score = ScoreWidget(window)
        self.mf = MoleculeFactory()
        self.blend = pyglet.resource.image("blend.png")
        self.heart = pyglet.resource.image("heart.png")
        self.eat = pyglet.resource.media('eat.wav', streaming=False)
        self.explode = pyglet.resource.media('explode.wav', streaming=False)
        self.cell = Cell(500,300)
        self.cell_emitter = PositionalBubbleEmitter(window, self.cell)
        self.game_map = GameMap("level1")
        self.camera = Camera(self.cell, self.window, self.game_map)
        self.life = 9
        self.fps_display = pyglet.clock.ClockDisplay()

    def update(self, dt):
        self.cell_emitter.update(dt)
        self.cell.update(dt, self.game_map)
        self.update_molecules(dt)
        self.update_camera(dt)
        
        # game over
        if self.life <= 0:
            self.window.replace_state(GameOver, self)
        
        
    def update_molecules(self, dt):
        self.game_map.spawner.update(dt)
        for m in self.game_map.spawner.molecules:
            if self.cell.collides_with(m):
                if sum(m.numbers) == 9:
                    self.eat.play().volume = 0.1
                    self.score.add(300)
                elif sum(m.numbers) == 18:
                    self.eat.play().volume = 0.15
                    self.score.add(1000)
                else:
                    self.explode.play().volume = 0.1
                    self.life -= 1
                m.kill()
                
            m.update(dt, self.game_map)

    def update_camera(self, dt):
        self.camera.calc_offset(dt)

    def on_draw(self):
        self.window.clear()
        glLoadIdentity()
        glPushMatrix()
        ## move the projection to the cameras offset
        offset = self.camera.get_offset()
        glTranslatef(offset[0], offset[1], 0.0)
        self.game_map.draw()
        self.cell_emitter.draw()
        ## offset the blend to always be on the screen
        self.blend.blit(abs(offset[0]),abs(offset[1]))
        ## Draw player
        self.cell.draw()
        for m in self.game_map.spawner.molecules:
            m.draw()
        glPopMatrix()
        ## Draw the ui stuffs
        self.score.draw()
        #self.fps_display.draw()
        heart_x_offset = 20
        for i in range(self.life):
            self.heart.blit(heart_x_offset, self.window.height-50)
            heart_x_offset += 45
        return pyglet.event.EVENT_HANDLED

    def on_key_press(self, k, args):
        ## handle arrow key movement
        if k == key.UP:
            self.cell.moving.y = 1
        elif k == key.DOWN:
            self.cell.moving.y = -1
        elif k == key.LEFT:
            self.cell.moving.x = -1
        elif k == key.RIGHT:
            self.cell.moving.x = 1
            
        ## Turn on debugging mode
        elif k == key.D:
            if self.game_map.debug == False:
                self.game_map.debug = True
            else:
                self.game_map.debug = False
        ## exit the current game
        elif k == key.ESCAPE:
            self.window.pop_state()
    
        ## pause the game
        elif k == key.ENTER:
            self.window.push_state(Pause, self)

        return pyglet.event.EVENT_HANDLED
    
    def on_key_release(self, k, args):
        if k == key.UP or k == key.DOWN:
            self.cell.moving.y = 0
        elif k == key.LEFT or k == key.RIGHT:
            self.cell.moving.x = 0
            
        return pyglet.event.EVENT_HANDLED

    def on_gain_focus(self):
        pyglet.clock.schedule_interval(self.update, 1.0/60.0)
        self.music = pyglet.resource.media('play02.mp3')
        self.player = self.music.play()
        self.player.queue(pyglet.resource.media('play03.mp3'))
        self.player.queue(pyglet.resource.media('play01.mp3'))
        self.player.volume = 0.1
        return pyglet.event.EVENT_HANDLED
    
    def on_lose_focus(self):
        self.player.pause()
        pyglet.clock.unschedule(self.update)
        pyglet.clock.unschedule(self.update_molecules)
        pyglet.clock.unschedule(self.update_camera)
        return pyglet.event.EVENT_HANDLED


###############################################################################
################################################################  Pause   #####
###############################################################################
class Pause(State):
    def __init__(self, window, *args, **kwargs):
        State.__init__(self, window, *args, **kwargs)
        self.old_state = args[0]
        self.label = pyglet.text.Label(text='Paused', x = window.width // 2,
                                    y = window.width // 2,
                                    anchor_x = 'center',
                                    anchor_y = 'center',
                                    font_size = 46)
        
    def on_draw(self):
        self.old_state.on_draw()
        self.label.draw()
        return pyglet.event.EVENT_HANDLED
    
    def on_key_press(self, k, mod):
        self.window.pop_state()
        return pyglet.event.EVENT_HANDLED
    
###############################################################################
################################################################  GameOver#####
###############################################################################
class GameOver(State):
    def __init__(self, window, *args, **kwargs):
        State.__init__(self, window, *args, **kwargs)
        self.old_state = args[0]
        points = self.old_state.score.get_score()
        self.label1 = pyglet.text.Label(text='You scored a total of %d points!' % points,
                                        font_size=24, x=window.width//2, y=window.height - window.height//3,
                                        anchor_x = 'center', anchor_y = 'center')
        self.label2 = pyglet.text.Label(text='Press ENTER to return to menu.',
                                        font_size=24, x=window.width//2, y=window.height - window.height//3 - 50,
                                        anchor_x = 'center', anchor_y = 'center')                                        
    def on_draw(self):
        self.old_state.on_draw()
        self.label1.draw()
        self.label2.draw()
        return pyglet.event.EVENT_HANDLED
    
    def on_key_press(self, k, mod):
        if k == key.ENTER:
            self.window.pop_state()
        return pyglet.event.EVENT_HANDLED
 
